@@ -28,7 +28,11 @@ from hypothesis.internal.conjecture.floats import (
     float_to_lex,
     lex_to_float,
 )
-from hypothesis.internal.conjecture.junkdrawer import replace_all
+from hypothesis.internal.conjecture.junkdrawer import (
+    binary_search,
+    pop_random,
+    replace_all,
+)
 from hypothesis.internal.conjecture.shrinking import Float, Integer, Lexical, Ordering
 from hypothesis.internal.conjecture.shrinking.common import find_integer
 
@@ -97,12 +101,10 @@ def defines_shrink_pass(generate_arguments):
     """A convenient decorator for defining shrink passes."""
 
     def accept(run_step):
-        definition = ShrinkPassDefinition(
-            generate_arguments=generate_arguments, run_step=run_step
-        )
+        ShrinkPassDefinition(generate_arguments=generate_arguments, run_step=run_step)
 
-        def run(self):
-            return self.run_shrink_pass(definition.name)
+        def run(self):  # pragma: no cover
+            assert False, "Shrink passes should not be run directly"
 
         run.__name__ = run_step.__name__
         run.is_shrink_pass = True
@@ -277,6 +279,7 @@ class Shrinker(object):
         self.__predicate = predicate
         self.__shrinking_prefixes = set()
         self.__derived_values = {}
+        self.__pending_shrink_explanation = None
 
         self.initial_size = len(initial.buffer)
 
@@ -291,6 +294,12 @@ class Shrinker(object):
         self.passes_by_name = {}
         self.passes = []
 
+    def explain_next_call_as(self, explanation):
+        self.__pending_shrink_explanation = explanation
+
+    def clear_call_explanation(self):
+        self.__pending_shrink_explanation = None
+
     def add_new_pass(self, run):
         """Creates a shrink pass corresponding to calling ``run(self)``"""
 
@@ -308,6 +317,8 @@ class Shrinker(object):
 
     def shrink_pass(self, name):
         """Return the ShrinkPass object for the pass with the given name."""
+        if isinstance(name, ShrinkPass):
+            return name
         if name not in self.passes_by_name:
             self.add_new_pass(name)
         return self.passes_by_name[name]
@@ -366,6 +377,10 @@ class Shrinker(object):
         too short to be a valid test case) or a ConjectureData object
         with status >= INVALID that would result from running this buffer."""
 
+        if self.__pending_shrink_explanation is not None:
+            self.debug(self.__pending_shrink_explanation)
+            self.__pending_shrink_explanation = None
+
         buffer = hbytes(buffer)
         result = self.__engine.cached_test_function(buffer)
         self.incorporate_test_data(result)
@@ -378,27 +393,6 @@ class Shrinker(object):
     def random(self):
         return self.__engine.random
 
-    def run_shrink_pass(self, sp):
-        """Runs the function associated with ShrinkPass sp and updates the
-        relevant metadata.
-
-        Note that sp may or may not be a pass currently associated with
-        this shrinker. This does not handle any requeing that is
-        required.
-        """
-        sp = self.shrink_pass(sp)
-
-        self.debug("Shrink Pass %s" % (sp.name,))
-        try:
-            sp.runs += 1
-
-            steps = sp.generate_steps()
-            self.random.shuffle(steps)
-            for s in steps:
-                sp.run_step(s)
-        finally:
-            self.debug("Shrink Pass %s completed." % (sp.name,))
-
     def shrink(self):
         """Run the full set of shrinks and update shrink_target.
 
@@ -489,45 +483,25 @@ class Shrinker(object):
         it twice will have exactly the same effect as calling it once.
         """
 
-        # "coarse" passes are ones which either make large scale modifications
-        # to the test case (alphabet_minimize) or delete data from it (the
-        # rest). After these have reached a fixed point the test case should
-        # be reasonably small and well normalized.
-        coarse = [
-            "alphabet_minimize",
-            "pass_to_descendant",
-            "zero_examples",
-            "adaptive_example_deletion",
-        ]
-        self.fixate_shrink_passes(coarse)
-
-        # "fine" passes are ones that make lots of fine grained changes
-        # to the shrink target. Typically we might expect these to make many
-        # small changes, with some changes unlocking others. Additionally,
-        # if there's more data left to delete then often running these is
-        # wasteful. As a result we only start running them after we've hit
-        # a fixed point for the coarse passes at least once.
-        fine = [
-            "reorder_examples",
-            "minimize_floats",
-            "minimize_duplicated_blocks",
-            "minimize_individual_blocks",
-        ]
-
-        self.fixate_shrink_passes(coarse + fine)
-
-        # "emergency" shrink passes are ones that handle cases that we
-        # can't currently handle more elegantly - either they're slightly
-        # weird hacks that happen to work or they're expensive passes to
-        # run. Generally we hope that the emergency passes don't do anything
-        # at all.
-        emergency = [
-            block_program("-XX"),
-            block_program("XX"),
-            "example_deletion_with_block_lowering",
-        ]
-
-        self.fixate_shrink_passes(coarse + fine + emergency)
+        self.fixate_shrink_passes(
+            [
+                block_program("X" * 5),
+                block_program("X" * 4),
+                block_program("X" * 3),
+                block_program("X" * 2),
+                block_program("X" * 1),
+                "pass_to_descendant",
+                "adaptive_example_deletion",
+                "alphabet_minimize",
+                "zero_examples",
+                "reorder_examples",
+                "minimize_floats",
+                "minimize_duplicated_blocks",
+                "minimize_individual_blocks",
+                block_program("-XX"),
+                "example_deletion_with_block_lowering",
+            ]
+        )
 
     def fixate_shrink_passes(self, passes):
         """Run steps from each pass in ``passes`` until the current shrink target
@@ -538,13 +512,7 @@ class Shrinker(object):
         while initial is not self.shrink_target:
             initial = self.shrink_target
             for sp in passes:
-                if sp.arguments:
-                    sp.runs += 1
-
-            passes_with_steps = [
-                (sp, step) for sp in passes for step in sp.generate_steps()
-            ]
-            self.random.shuffle(passes_with_steps)
+                sp.runs += 1
 
             # We run remove_discarded after every step to do cleanup
             # keeping track of whether that actually works. Either there is
@@ -554,10 +522,91 @@ class Shrinker(object):
             # try again once all of the passes have been run.
             can_discard = self.remove_discarded()
 
-            for sp, step in passes_with_steps:
-                sp.run_step(step)
-                if can_discard:
-                    can_discard &= self.remove_discarded()
+            # We want to run all steps from each pass, but the order in which
+            # we do this makes a fairly significant difference. If a pass has
+            # a lot of steps but is useless, we'd like to defer running those
+            # steps (many of which will become faster or irrelevant) until
+            # after other, better, passes have had a chance to run.
+            #
+            # An ideal outcome would be that we take every pass/step pair,
+            # run every one of these pairs that would produce a successful
+            # shrink, then run every pair that wouldn't.
+            #
+            # Doing this would of course require us to be magically good at
+            # predicting which pairs would work, and if we could do that then
+            # we wouldn't bother running the unsuccessful ones at all. Instead
+            # what we do is we iterate over them in an order that prioritises
+            # passes that seem to work well and penalises ones that make useless
+            # test calls.
+
+            # The passes we've not run to completion yet, along with the steps
+            # they've yet to run. The steps start out as None, so that we only
+            # calculate the list the first time we run the pass. This is because
+            # if the shrink target changes between calculating steps and running
+            # them, a lot of the steps become invalid. Particularly when the
+            # initial shrink target is large and thus there are a lot of steps
+            # this is potentially fairly wasteful, so we defer calculation to
+            # point of first usage where they will definitely be valid.
+            passes_with_steps = [(sp, None) for sp in passes]
+
+            # Run passes that are already at a fixed point last, so that we've
+            # had a chance to unlock them by the time we come to run them.
+            passes_with_steps.sort(
+                key=lambda t: t[0].fixed_point_at is self.shrink_target
+            )
+            successful_passes = set()
+
+            while passes_with_steps:
+                to_run_next = []
+
+                for sp, steps in passes_with_steps:
+                    if steps is None:
+                        steps = sp.generate_steps()
+
+                    # We run each pass until it has failed a certain number
+                    # of times, where a "failure" is any step where it made
+                    # at least one call and did not result in a shrink.
+                    # This gives passes which work reasonably often more of
+                    # chance to run.
+                    failures = 0
+                    successes = 0
+
+                    # The choice of 3 is fairly arbitrary and was hand tuned
+                    # to some particular examples. It is very unlikely that
+                    # is the best choice in general, but it's not an
+                    # unreasonable choice: Making it smaller than this would
+                    # give too high a chance of an otherwise very worthwhile
+                    # pass getting screened out too early if it got unlucky,
+                    # and making it much larger than this would result in us
+                    # spending too much time on bad passes.
+                    max_failures = 3
+
+                    while steps and failures < max_failures:
+                        prev_calls = self.calls
+                        prev = self.shrink_target
+                        sp.run_step(pop_random(self.random, steps))
+                        if prev_calls != self.calls:
+                            if can_discard:
+                                can_discard = self.remove_discarded()
+                            if prev is self.shrink_target:
+                                failures += 1
+                            else:
+                                successes += 1
+                    if steps:
+                        to_run_next.append((sp, steps))
+                    if successes > 0:
+                        successful_passes.add(sp)
+
+                passes_with_steps = to_run_next
+
+            # If only some of our shrink passes are doing anything useful
+            # then run all of those to a fixed point before running the
+            # full set. This is particularly important when an emergency
+            # shrink pass unlocks some non-emergency ones and it suddenly
+            # becomes very expensive to find a bunch of small changes.
+            if 0 < len(successful_passes) < len(passes):
+                self.fixate_shrink_passes(successful_passes)
+
         for sp in passes:
             sp.fixed_point_at = self.shrink_target
 
@@ -573,10 +622,8 @@ class Shrinker(object):
     def examples(self):
         return self.shrink_target.examples
 
-    def all_block_bounds(self, target=None):
-        if target is None:
-            target = self.shrink_target
-        return [(b.start, b.end) for b in target.blocks]
+    def all_block_bounds(self):
+        return self.shrink_target.blocks.all_bounds()
 
     @derived_value
     def examples_by_label(self):
@@ -610,11 +657,17 @@ class Shrinker(object):
                         hi = mid
                     else:
                         lo = mid
-                result.extend([(ex, ls[j]) for j in hrange(i + 1, hi)])
+                id1 = self.stable_identifier_for_example(ex)
+                result.extend(
+                    [
+                        (id1, self.stable_identifier_for_example(ls[j]))
+                        for j in hrange(i + 1, hi)
+                    ]
+                )
         return result
 
     @defines_shrink_pass(calculate_descents)
-    def pass_to_descendant(self, ancestor, descendant):
+    def pass_to_descendant(self, ancestor_id, descendant_id):
         """Attempt to replace each example with a descendant example.
 
         This is designed to deal with strategies that call themselves
@@ -633,6 +686,12 @@ class Shrinker(object):
         as possible.
         """
 
+        ancestor = self.example_for_stable_identifier(ancestor_id)
+        descendant = self.example_for_stable_identifier(descendant_id)
+        if descendant is None:
+            return
+        assert ancestor is not None
+
         self.incorporate_new_buffer(
             self.buffer[: ancestor.start]
             + self.buffer[descendant.start : descendant.end]
@@ -696,7 +755,7 @@ class Shrinker(object):
 
         current = self.shrink_target
 
-        blocked = [current.buffer[u:v] for u, v in self.all_block_bounds(current)]
+        blocked = [current.buffer[u:v] for u, v in self.all_block_bounds()]
 
         changed = [
             i
@@ -735,29 +794,66 @@ class Shrinker(object):
             self.__shrinking_prefixes.add(prefix)
 
     def clear_change_tracking(self):
-        self.__changed_blocks.clear()
+        self.__last_checked_changed_at = self.shrink_target
+        self.__all_changed_blocks = set()
 
     def mark_changed(self, i):
         self.__changed_blocks.add(i)
 
+    @property
+    def __changed_blocks(self):
+        if self.__last_checked_changed_at is not self.shrink_target:
+            prev_target = self.__last_checked_changed_at
+            new_target = self.shrink_target
+            assert prev_target is not new_target
+            prev = prev_target.buffer
+            new = new_target.buffer
+            assert sort_key(new) < sort_key(prev)
+
+            if (
+                len(new_target.blocks) != len(prev_target.blocks)
+                or new_target.blocks.endpoints != prev_target.blocks.endpoints
+            ):
+                self.__all_changed_blocks = set()
+            else:
+                blocks = new_target.blocks
+
+                # Index of last block whose contents have been modified, found
+                # by checking if the tail past this point has been modified.
+                last_changed = binary_search(
+                    0,
+                    len(blocks),
+                    lambda i: prev[blocks.start(i) :] != new[blocks.start(i) :],
+                )
+
+                # Index of the first block whose contents have been changed,
+                # because we know that this predicate is true for zero (because
+                # the prefix from the start is empty), so the result must be True
+                # for the bytes from the start of this block and False for the
+                # bytes from the end, hence the change is in this block.
+                first_changed = binary_search(
+                    0,
+                    len(blocks),
+                    lambda i: prev[: blocks.start(i)] == new[: blocks.start(i)],
+                )
+
+                # Between these two changed regions we now do a linear scan to
+                # check if any specific block values have changed.
+                for i in hrange(first_changed, last_changed + 1):
+                    u, v = blocks.bounds(i)
+                    if i not in self.__all_changed_blocks and prev[u:v] != new[u:v]:
+                        self.__all_changed_blocks.add(i)
+            self.__last_checked_changed_at = new_target
+        assert self.__last_checked_changed_at is self.shrink_target
+        return self.__all_changed_blocks
+
     def update_shrink_target(self, new_target):
         assert isinstance(new_target, ConjectureResult)
         if self.shrink_target is not None:
-            current = self.shrink_target.buffer
-            new = new_target.buffer
-            assert sort_key(new) < sort_key(current)
             self.shrinks += 1
-            if (
-                len(new_target.blocks) != len(self.shrink_target.blocks)
-                or self.all_block_bounds(new_target) != self.all_block_bounds()
-            ):
-                self.clear_change_tracking()
-            else:
-                for i, (u, v) in enumerate(self.all_block_bounds()):
-                    if i not in self.__changed_blocks and current[u:v] != new[u:v]:
-                        self.mark_changed(i)
         else:
-            self.__changed_blocks = set()
+            self.__all_changed_blocks = set()
+            self.__last_checked_changed_at = new_target
 
         self.shrink_target = new_target
         self.__shrinking_block_cache = {}
@@ -919,36 +1015,11 @@ class Shrinker(object):
         return True
 
     @derived_value
-    def endpoints_by_depth(self):
-        """Defines a series of increasingly fine grained boundaries
-        to partition the current buffer, based on the depth of examples.
-        Each element of the result is the set of endpoints of examples
-        less than or equal to some depth (less than or equal to account
-        for the fact that there might be blocks below that depth and if
-        we ignore those we'll get bad boundaries).
+    def stable_identifier_arguments(self):
+        return [(self.stable_identifier_for_example(ex),) for ex in self.examples]
 
-        This is primarily useful for adaptive_example_deletion."""
-        endpoints_at_depth = defaultdict(set)
-        max_depth = 0
-        for ex in self.examples:
-            endpoints_at_depth[ex.depth].add(ex.start)
-            endpoints_at_depth[ex.depth].add(ex.end)
-            max_depth = max(max_depth, ex.depth)
-        distinct_partitions = [{0, len(self.buffer)}]
-        for d in hrange(max_depth + 1):
-            prev = distinct_partitions[-1]
-            if not endpoints_at_depth[d].issubset(prev):
-                distinct_partitions.append(prev | endpoints_at_depth[d])
-        return [sorted(endpoints) for endpoints in distinct_partitions[1:]]
-
-    @defines_shrink_pass(
-        lambda self: [
-            (i, j)
-            for i, ls in enumerate(self.endpoints_by_depth)
-            for j in hrange(len(ls))
-        ]
-    )
-    def adaptive_example_deletion(self, i, j):
+    @defines_shrink_pass(lambda self: self.stable_identifier_arguments)
+    def adaptive_example_deletion(self, example_id):
         """Attempts to delete every example from the test case.
 
         That is, it is logically equivalent to trying ``self.buffer[:ex.start] +
@@ -956,18 +1027,32 @@ class Shrinker(object):
         examples are tried is randomized, and when deletion is successful it
         will attempt to adapt to delete more than one example at a time.
         """
-        partition = self.endpoints_by_depth[i]
-        # No point in trying to delete the last element because that will always
-        # give us a prefix.
-        if j >= len(partition) - 1:
+        example = self.example_for_stable_identifier(example_id)
+
+        if example is None or not self.incorporate_new_buffer(
+            self.buffer[: example.start] + self.buffer[example.end :]
+        ):
             return
 
+        # If we successfully deleted one example there may be a useful
+        # deletable region around here.
+
+        original = self.shrink_target
+        endpoints = set()
+        for ex in original.examples:
+            if ex.depth <= example.depth:
+                endpoints.add(ex.start)
+                endpoints.add(ex.end)
+
+        partition = sorted(endpoints)
+        j = partition.index(example.start)
+
         def delete_region(a, b):
             assert a <= j <= b
             if a < 0 or b >= len(partition) - 1:
                 return False
             return self.consider_new_buffer(
-                self.buffer[: partition[a]] + self.buffer[partition[b] :]
+                original.buffer[: partition[a]] + original.buffer[partition[b] :]
             )
 
         to_right = find_integer(lambda n: delete_region(j, j + n))
@@ -975,8 +1060,7 @@ class Shrinker(object):
         if to_right > 0:
             find_integer(lambda n: delete_region(j - n, j + to_right))
 
-    def try_zero_example(self, i):
-        ex = self.examples[i]
+    def try_zero_example(self, ex):
         u = ex.start
         v = ex.end
         attempt = self.cached_test_function(
@@ -994,22 +1078,25 @@ class Shrinker(object):
                 self.incorporate_new_buffer(
                     self.buffer[:u] + hbytes(used) + self.buffer[v:]
                 )
-        return self.examples[i].trivial
+        return self.examples[ex.index].trivial
 
-    @defines_shrink_pass(lambda self: [(e.index,) for e in self.examples])
-    def zero_examples(self, ex_index):
+    @defines_shrink_pass(lambda self: self.stable_identifier_arguments)
+    def zero_examples(self, ex_id):
+        """Attempt to replace each example with a minimal version of itself."""
+        ex = self.example_for_stable_identifier(ex_id)
         """Attempt to replace each example with a minimal version of itself."""
         # If the example is already trivial, assume there's nothing to do here.
         # We could attempt to use it as an adaptive replacement for other
         # similar examples, but that seems to be ineffective, resulting mostly
         # in redundant work rather than helping.
-        if self.examples[ex_index].trivial:
+        if ex is None or ex.trivial:
             return
 
-        if not self.try_zero_example(ex_index):
+        if not self.try_zero_example(ex):
             return
 
-        ex = self.examples[ex_index]
+        # If we zeroed the example we need to get the new one that replaced it.
+        ex = self.examples[ex.index]
 
         original = self.shrink_target
         group = self.examples_by_label[ex.label]
@@ -1045,7 +1132,8 @@ class Shrinker(object):
 
         expand_region(can_zero, start, end)
 
-    def duplicated_block_suffixes(self):
+    @derived_value
+    def blocks_by_non_zero_suffix(self):
         """Returns a list of blocks grouped by their non-zero suffix,
         as a list of (suffix, indices) pairs, skipping all groupings
         where there is only one index.
@@ -1057,16 +1145,12 @@ class Shrinker(object):
             duplicates[non_zero_suffix(self.buffer[block.start : block.end])].append(
                 block.index
             )
-        return sorted(
-            [
-                (suffix, indices)
-                for suffix, indices in duplicates.items()
-                if len(suffix) > 0 and len(indices) > 1
-            ]
-        )
+        return duplicates
 
-    @defines_shrink_pass(duplicated_block_suffixes)
-    def minimize_duplicated_blocks(self, block, targets):
+    @defines_shrink_pass(
+        lambda self: [(b,) for b in sorted(self.blocks_by_non_zero_suffix)]
+    )
+    def minimize_duplicated_blocks(self, block):
         """Find blocks that have been duplicated in multiple places and attempt
         to minimize all of the duplicates simultaneously.
 
@@ -1086,6 +1170,9 @@ class Shrinker(object):
         of the blocks doesn't matter very much because it allows us to replace
         more values at once.
         """
+        targets = self.blocks_by_non_zero_suffix[block]
+        if len(targets) <= 1:
+            return
         Lexical.shrink(
             block,
             lambda b: self.try_shrinking_blocks(targets, b),
@@ -1093,18 +1180,8 @@ class Shrinker(object):
             full=False,
         )
 
-    @defines_shrink_pass(
-        lambda self: [
-            (ex,)
-            for ex in self.shrink_target.examples
-            if (
-                ex.label == DRAW_FLOAT_LABEL
-                and len(ex.children) == 2
-                and ex.children[0].length == 8
-            )
-        ]
-    )
-    def minimize_floats(self, ex):
+    @defines_shrink_pass(lambda self: self.stable_identifier_arguments)
+    def minimize_floats(self, ex_id):
         """Some shrinks that we employ that only really make sense for our
         specific floating point encoding that are hard to discover from any
         sort of reasonable general principle. This allows us to make
@@ -1120,6 +1197,13 @@ class Shrinker(object):
         transformations to make, they just don't necessarily correspond to
         anything particularly meaningful for non-float values.
         """
+        ex = self.example_for_stable_identifier(ex_id)
+        if ex is None or not (
+            ex.label == DRAW_FLOAT_LABEL
+            and len(ex.children) == 2
+            and ex.children[0].length == 8
+        ):
+            return
         u = ex.children[0].start
         v = ex.children[0].end
         buf = self.shrink_target.buffer
@@ -1137,8 +1221,8 @@ class Shrinker(object):
                 random=self.random,
             )
 
-    @defines_shrink_pass(lambda self: [(b,) for b in self.blocks])
-    def minimize_individual_blocks(self, block):
+    @defines_shrink_pass(lambda self: [(i,) for i in hrange(len(self.blocks))])
+    def minimize_individual_blocks(self, i):
         """Attempt to minimize each block in sequence.
 
         This is the pass that ensures that e.g. each integer we draw is a
@@ -1149,6 +1233,9 @@ class Shrinker(object):
 
         then in our shrunk example, x = 10 rather than say 97.
         """
+        if i >= len(self.blocks):
+            return
+        block = self.blocks[i]
         u, v = block.bounds
         i = block.index
         Lexical.shrink(
@@ -1160,14 +1247,14 @@ class Shrinker(object):
 
     @defines_shrink_pass(
         lambda self: [
-            (block, ex)
-            for block in self.blocks
-            if self.is_shrinking_block(block.index)
+            (i, self.stable_identifier_for_example(ex))
+            for i in hrange(len(self.blocks))
+            if self.is_shrinking_block(i)
             for ex in self.examples
-            if ex.start >= block.end and ex.length > 0
+            if ex.length > 0 and ex.start >= self.blocks[i].end
         ]
     )
-    def example_deletion_with_block_lowering(self, block, ex):
+    def example_deletion_with_block_lowering(self, block_i, ex_id):
         """Sometimes we get stuck where there is data that we could easily
         delete, but it changes the number of examples generated, so we have to
         change that at the same time.
@@ -1181,6 +1268,12 @@ class Shrinker(object):
         example and every block not inside that example it tries deleting the
         example and modifying the block's value by one in either direction.
         """
+        if block_i >= len(self.blocks):
+            return
+        block = self.blocks[block_i]
+        ex = self.example_for_stable_identifier(ex_id)
+        if ex is None:
+            return
         u, v = block.bounds
 
         n = int_from_bytes(self.shrink_target.buffer[u:v])
@@ -1192,19 +1285,14 @@ class Shrinker(object):
         del buf[ex.start : ex.end]
         self.incorporate_new_buffer(buf)
 
-    @derived_value
-    def groups_for_reordering(self):
-        result = defaultdict(list)
-        for e in self.examples:
-            result[(e.parent, e.label)].append(e)
-        groups = [v for v in result.values() if len(v) > 1]
-        groups.sort(key=lambda ls: ls[0].index)
-        return groups
-
     @defines_shrink_pass(
-        lambda self: [(i,) for i in hrange(len(self.groups_for_reordering))]
+        lambda self: [
+            (self.stable_identifier_for_example(e), l)
+            for e in self.examples
+            for l in sorted({c.label for c in e.children}, key=str)
+        ]
     )
-    def reorder_examples(self, i):
+    def reorder_examples(self, ex_id, label):
         """This pass allows us to reorder the children of each example.
 
         For example, consider the following:
@@ -1222,7 +1310,13 @@ class Shrinker(object):
         ``x=""``, ``y="0"``, or the other way around. With reordering it will
         reliably fail with ``x=""``, ``y="0"``.
         """
-        group = self.groups_for_reordering[i]
+        ex = self.example_for_stable_identifier(ex_id)
+        if ex is None:
+            return
+        group = [c for c in ex.children if c.label == label]
+        if len(group) <= 1:
+            return
+
         st = self.shrink_target
         pieces = [st.buffer[ex.start : ex.end] for ex in group]
         endpoints = [(ex.start, ex.end) for ex in group]
@@ -1309,6 +1403,20 @@ class Shrinker(object):
             else:
                 lo = mid
 
+    def stable_identifier_for_example(self, example):
+        """A stable identifier is one that we can reasonably reliably
+        count on referring to "logically the same" example between two
+        different test runs. It is currently represented as a path from
+        the root."""
+        return example.index
+
+    def example_for_stable_identifier(self, identifier):
+        """Returns the example in the current shrink target corresponding
+        to this stable identifier, or None if no such example exists."""
+        if identifier >= len(self.examples):
+            return None
+        return self.shrink_target.examples[identifier]
+
     def run_block_program(self, i, description, original, repeats=1):
         """Block programs are a mini-DSL for block rewriting, defined as a sequence
         of commands that can be run at some index into the blocks
@@ -1372,13 +1480,14 @@ def block_program(description):
         n = len(description)
 
         def generate_arguments(self):
-            return [(i,) for i in hrange(len(self.blocks) - n)]
+            return [(i,) for i in hrange(len(self.blocks))]
 
         def run(self, i):
             """Adaptively attempt to run the block program at the current
             index. If this successfully applies the block program ``k`` times
             then this runs in ``O(log(k))`` test function calls."""
-            assert i + n < len(self.shrink_target.blocks)
+            if i + n >= len(self.shrink_target.blocks):
+                return
 
             # First, run the block program at the chosen index. If this fails,
             # don't do any extra work, so that failure is as cheap as possible.
@@ -1423,9 +1532,6 @@ class ShrinkPass(object):
     index = attr.ib()
     shrinker = attr.ib()
 
-    __arguments = attr.ib(default=None, init=False)
-    __target_at_argument_calculation = attr.ib(default=None, init=False)
-
     fixed_point_at = attr.ib(default=None)
     successes = attr.ib(default=0)
     runs = attr.ib(default=0)
@@ -1433,32 +1539,25 @@ class ShrinkPass(object):
     shrinks = attr.ib(default=0)
     deletions = attr.ib(default=0)
 
-    @property
-    def arguments(self):
-        if self.__target_at_argument_calculation is not self.shrinker.shrink_target:
-            self.__arguments = self.generate_arguments(self.shrinker)
-            self.__target_at_argument_calculation = self.shrinker.shrink_target
-        return self.__arguments
-
     def generate_steps(self):
         if self.fixed_point_at is self.shrinker.shrink_target:
             return []
-        return list(hrange(len(self.arguments)))
+        return list(self.generate_arguments(self.shrinker))
 
-    def run_step(self, i):
-        if i >= len(self.arguments):
-            return
-        args = self.arguments[i]
-        self.shrinker.debug("%s(%s)" % (self.name, ", ".join(map(repr, args))))
+    def run_step(self, args):
         initial_shrinks = self.shrinker.shrinks
         initial_calls = self.shrinker.calls
         size = len(self.shrinker.shrink_target.buffer)
+        self.shrinker.explain_next_call_as(
+            "%s(%s)" % (self.name, ", ".join(map(repr, args)))
+        )
         try:
             self.run_with_arguments(self.shrinker, *args)
         finally:
             self.calls += self.shrinker.calls - initial_calls
             self.shrinks += self.shrinker.shrinks - initial_shrinks
             self.deletions += size - len(self.shrinker.shrink_target.buffer)
+            self.shrinker.clear_call_explanation()
 
     @property
     def name(self):
